@@ -2,12 +2,13 @@
 #![cfg_attr(not(any(doc, test, feature = "std")), no_std)]
 
 use core::ops::{Add, AddAssign, Div, DivAssign, Mul, MulAssign, Sub, SubAssign};
+use std::mem::{needs_drop, MaybeUninit};
 
 #[cfg(test)]
 mod tests;
 
 mod iter;
-use iter::{BinOpsIter, Iter};
+use iter::Slice;
 
 pub struct Array<T, const N: usize>(pub [T; N]);
 
@@ -16,17 +17,57 @@ fn binop_impl<T, U, O, const N: usize>(
     rhs: [U; N],
     op: impl Fn(T, U) -> O + Copy,
 ) -> [O; N] {
-    let mut dc = BinOpsIter::new(lhs, rhs);
+    if !needs_drop::<T>() && !needs_drop::<U>() && !needs_drop::<O>() {
+        // SAFETY:
+        // we've just checked that T, U and O are non-drop types
+        unsafe { binop_impl_copy(lhs, rhs, op) }
+    } else {
+        binop_impl_drop(lhs, rhs, op)
+    }
+}
+
+fn binop_impl_drop<T, U, O, const N: usize>(
+    lhs: [T; N],
+    rhs: [U; N],
+    op: impl Fn(T, U) -> O + Copy,
+) -> [O; N] {
+    let mut lhs = Slice::full(lhs);
+    let mut rhs = Slice::full(rhs);
+    let mut output = Slice::new();
 
     for _ in 0..N {
-        // SAFETY:
-        // Will only be called a maximum of N times
-        unsafe { dc.step(op) }
+        unsafe {
+            let lhs = lhs.pop_front_unchecked();
+            let rhs = rhs.pop_front_unchecked();
+            output.push_unchecked(op(lhs, rhs));
+        }
     }
 
+    unsafe { output.output() }
+}
+
+/// # Safety
+/// must only be called if T, U and O are Copy types (no drop needed)
+unsafe fn binop_impl_copy<T, U, O, const N: usize>(
+    lhs: [T; N],
+    rhs: [U; N],
+    op: impl Fn(T, U) -> O + Copy,
+) -> [O; N] {
     // SAFETY:
-    // By this point, we are certain we have initialised all N elements
-    unsafe { dc.output() }
+    // we will not read from output, and caller ensures that O is non-drop
+    // this makes miri sad by saves a memcpy on the return value so idc
+    #[allow(clippy::uninit_assumed_init)]
+    let mut output: [O; N] = unsafe { MaybeUninit::uninit().assume_init() };
+
+    for i in 0..N {
+        unsafe {
+            let lhs = core::ptr::read(&lhs[i]);
+            let rhs = core::ptr::read(&rhs[i]);
+            output[i] = op(lhs, rhs);
+        }
+    }
+
+    output
 }
 
 fn binop_assign_impl<T, U, const N: usize>(
@@ -34,12 +75,12 @@ fn binop_assign_impl<T, U, const N: usize>(
     rhs: [U; N],
     op: impl Fn(&mut T, U) + Copy,
 ) {
-    let mut dc = Iter::new(rhs);
+    let mut rhs = Slice::full(rhs);
 
-    for _ in 0..N {
+    for i in 0..N {
         // SAFETY:
         // Will only be called a maximum of N times
-        unsafe { op(lhs.get_unchecked_mut(dc.index()), dc.next_unchecked()) }
+        unsafe { op(lhs.get_unchecked_mut(i), rhs.pop_front_unchecked()) }
     }
 }
 
